class ActivityAction:
    @staticmethod
    def create_activity():
        raise NotImplementedError

    @staticmethod
    def add_activity():
        raise NotImplementedError

    @staticmethod
    def list_activities():
        raise NotImplementedError

    @staticmethod
    def modify_activity():
        raise NotImplementedError

    @staticmethod
    def delete_activity():
        raise NotImplementedError

    @staticmethod
    def export_activities():
        raise NotImplementedError



from sport_tracker.controller.user_actions import UserActions
from sport_tracker.controller.activity_actions import ActivityAction
from sport_tracker.controller.sport_actions import SportAction

def process_args(arguments):
    actions_dict : dict = {
        'create-user': UserActions.create_user,
        'modify-user': UserActions.modify_user,
        'list-users': UserActions.list_users,
        'delete-user': UserActions.delete_user,
        'export-users': UserActions.export_users,
        'add-activity': ActivityAction.add_activity,
        'create-activity': ActivityAction.create_activity,
        'list-activities': ActivityAction.list_activities,
        'modify-activity': ActivityAction.modify_activity,
        'delete-activity': ActivityAction.delete_activity,
        'export-activities': ActivityAction.export_activities,
        'add-sport': SportAction.add_sport,
        'list-sports': SportAction.list_sports,
        'modify-sports': SportAction.modify_sport,
        'delete-sport': SportAction.delete_sport
    }

    # as arguments are checked by argparser, it's safe to execute args without
    # other checks
    actions_dict[arguments.action[0]](*arguments.parameters)



from sport_tracker.argparser import init_args
from sport_tracker.controller.argument_actions import process_args


def main():
    process_args(init_args())
 
    
if __name__ == '__main__':
    main()




from texttable import Texttable
from sport_tracker.common.exceptions import InvalidArgumentError


class TerminalOutput:
    @staticmethod
    def print_table(header: list, data: list):
        if len(header) != len(data[0]):
            raise InvalidArgumentError("Header and data must have the same number of columns")
        table = Texttable()
        table.add_rows([header] + data)
        print(table.draw())



from argparse import ArgumentParser

arg_parser = ArgumentParser()

user_actions = ['create-user',
                'modify-user',
                'list-users',
                'delete-user',
                'export-users']

activity_actions = ['add-activity',
                    'list-activities',
                    'modify-activity',
                    'delete-activity',
                    'export-activities']

sport_actions = ['add-sport',
                 'list-sports',
                 'modify-sports',
                 'delete-sport']

arg_parser.add_argument("action",
                        nargs=1,
                        help="SportTracker action",
                        choices=user_actions + activity_actions + sport_actions)
arg_parser.add_argument("parameters",
                        nargs="*",
                        help="Parameters of specified action")
arg_parser.add_argument("--no_confirm",
                        help="Disable confirmation dialogs",
                        action="store_true",
                        default=False)


def init_args():
    return arg_parser.parse_args()



from setuptools import setup

with open("requirements.txt") as req_file:
    REQUIREMENTS: list = req_file.readlines()

setup(
    name='sport_tracker',
    version='0.0.3',
    author='Katarina Bulkova',
    author_email='bulkova.katarina@gmail.com',
    description='CLI Sport Diary',
    packages=['sport_tracker', 'sport_tracker.common', 'sport_tracker.controller', 'sport_tracker.view'],
    platforms=['linux'],
    python_requires=">=3.7",
    license="GPL",
    entry_points={
        'console_scripts': [
            'sport_tracker = sport_tracker.__main__:main'
         ]
    },
    install_requires=REQUIREMENTS
)



from sport_tracker.controller.db_controller import DBController
from sport_tracker.logger import logger
from sport_tracker.view.terminal_output import TerminalOutput


class SportAction:
    @staticmethod
    def _obtain_sport_info(default_values: dict = None) -> (str, int):
        while True:
            # name input
            default_prompt = f" [{default_values['sport_name']}]" if default_values else ''
            sport_name: str = input(f"Input sport's name{default_prompt}: ")
            if sport_name:
                break
            if not sport_name and default_values:
                sport_name = default_values['sport_name']
                break
            else:
                logger.error("Sport's name cannot be empty")
        # moving input
        while True:
            default_prompt = f" [{default_values['moving']}]" if default_values else ''
            moving: str = input(f"Is this sport distance based (moving) {default_prompt}? ")
            if not moving and default_values:
                moving: int = default_values['moving']
                break
            if moving.lower().startswith('y'):
                moving: int = 1
                break
            elif moving.lower().startswith('n'):
                moving: int = 0
                break
            else:
                logger.error("Incorrect value, must be either yes or no")

        return sport_name, moving

    @staticmethod
    def add_sport():
        while True:
            logger.info("Registering new sport")
            sport_name, moving = SportAction._obtain_sport_info()
            # print information for user to verify
            print("\nVerify these information: ")
            print(f"Name of the sport: {sport_name}")
            print(f"Sport is distance based (moving): {'yes' if moving == 1 else 'no'}")
            while True:
                choice = input("Is this correct? [Y/n]: ") or "y"
                if choice.lower().startswith("y"):
                    correct = True
                    break
                elif choice.lower().startswith("n"):
                    correct = False
                    break

            if correct:
                break

                # insert obtained pieces of information into database
        with DBController() as db:
            db.insert_sport(name=sport_name, moving=moving==1)
            logger.info("Sport added")
            print(f"Sport {sport_name} added successfully.")

    @staticmethod
    def modify_sport():
        raise NotImplementedError

    @staticmethod
    def list_sports():
        with DBController() as db:
            TerminalOutput.print_table(header=['ID', 'Name', 'Distance based?'],
                                       data=db.fetch_all(table="sports"))

    @staticmethod
    def delete_sport():
        raise NotImplementedError


from datetime import date
from math import floor

from pint import UnitRegistry
from pint.errors import UndefinedUnitError, DimensionalityError

from sport_tracker.common.exceptions import InvalidArgumentError
from sport_tracker.controller.db_controller import DBController
from sport_tracker.logger import logger
from sport_tracker.model.person import ActivityLevel
from sport_tracker.view.terminal_output import TerminalOutput


class UserActions:
    @staticmethod
    def _obtain_user_info(default_values: dict = None):
        unit_registry = UnitRegistry()
        # name input
        while True:
            default_prompt = f" [{default_values['user_name']}]" if default_values else ''
            user_name: str = input(f"Input your name{default_prompt}: ")
            if user_name:
                break
            elif default_values:
                user_name = default_values['user_name']
                break
            else:
                logger.error("User's name cannot be empty")
        # birth date input
        while True:
            default_prompt = f" [{default_values['date_of_birth']}]" if default_values else ''
            date_of_birth: str = input(f"Input your birth date in format DD-MM-YYYY{default_prompt}: ")
            if not date_of_birth and default_values:
                date_of_birth = f"{default_values['date_of_birth'].split('-')[2]}-" \
                    f"{default_values['date_of_birth'].split('-')[1]}-" \
                    f"{default_values['date_of_birth'].split('-')[0]}"
            try:
                # TODO: Investigate this handling in comparison with name handling
                date_of_birth: date = date(day=int(date_of_birth.split("-")[0]),
                                           month=int(date_of_birth.split("-")[1]),
                                           year=int(date_of_birth.split("-")[2]))
                if not date_of_birth:
                    raise ValueError("Empty date")
            except (TypeError, IndexError):
                logger.error("Incorrect date format, try again")
            except ValueError:
                logger.error("Impossible date, try again")
            else:
                break
        # weight input
        while True:
            default_prompt = f" [{default_values['weight']}] kg" if default_values else ''
            weight: str = input(f"Enter your weight with units designation{default_prompt}: ")
            try:
                if not weight and default_values:
                    weight = f"{default_values['weight']} kg"
                elif not weight:
                    raise UndefinedUnitError
                weight: float = unit_registry.Quantity(weight).to(unit_registry.kilogram).magnitude
            except (UndefinedUnitError, DimensionalityError) as e:
                logger.error(str(e))
            else:
                break
        # height input
        while True:
            default_prompt = f" [{default_values['height']}] cm" if default_values else ''
            height: str = input(f"Enter your height with unit designation{default_prompt}: ")
            try:
                if not height and default_values:
                    height = f"{default_values['height']} cm"
                if not height:
                    raise UndefinedUnitError
                height: int = floor(unit_registry.Quantity(height).to(unit_registry.centimeter).magnitude)
            except (UndefinedUnitError, DimensionalityError) as e:
                logger.error(str(e))
            else:
                break
        # gender input
        while True:
            default_prompt = f" [{default_values['gender']}]" if default_values else ''
            gender: str = input(f"Enter your gender [male/female]{default_prompt}: ")
            if not gender and default_values:
                gender = default_values['gender']
                break
            if gender.lower().startswith("m"):
                gender: int = 0  # male is stored as 0 in db
                break
            elif gender.lower().startswith("f"):
                gender: int = 1  # female is stored as 1 in db
                break
            else:
                logger.error("Incorrect gender, use either male or female")
        # activity level input
        while True:
            default_prompt = f" [{default_values['activity_level']}]" if default_values else ''
            activity_level: str = input(f"Enter your activity level "
                                        f"[1 = sedentary to 4 = professional athlete]{default_prompt}: ")
            try:
                if not activity_level and default_values:
                    activity_level = default_values['activity_level']
                if not activity_level:
                    raise InvalidArgumentError
                activity_level: int = int(activity_level)
            except \
                    (TypeError, InvalidArgumentError):
                logger.error("Incorrect activity level, try again")
            else:
                if 1 <= activity_level <= 4:
                    break
                else:
                    logger.error("Activity level must be either 1, 2, 3 or 4")
        return user_name, date_of_birth, weight, height, gender, activity_level

    @staticmethod
    def create_user():
        unit_registry = UnitRegistry()
        while True:
            logger.info("Registering new user")
            user_name, date_of_birth, weight, height, gender, activity_level = UserActions._obtain_user_info()
            # print information for user to verify
            print("\nVerify these information: ")
            print(f"Name: {user_name}")
            print(f"Age: {floor((date.today() - date_of_birth).total_seconds() / 365 / 24 / 3600)} years")
            print(f"Weight: {weight * unit_registry.kilogram}")
            print(f"Height: {height * unit_registry.centimeter}")
            print(f"Gender: {'male' if gender == 0 else 'female'}")
            print(f"Activity level: {ActivityLevel(activity_level)}")

            while True:
                choice = input("Is this correct? [Y/n]: ") or "y"
                if choice.lower().startswith("y"):
                    correct = True
                    break
                elif choice.lower().startswith("n"):
                    correct = False
                    break

            if correct:
                break

        # insert obtained pieces of information into database
        with DBController() as db:
            db.insert_user(name=user_name, date_born=date_of_birth, weight=weight,
                           height=height, gender=gender, activity=ActivityLevel(activity_level))
            logger.info("User added")
            print(f"User {user_name} added successfully.")

    @staticmethod
    def list_users():
        with DBController() as db:
            TerminalOutput.print_table(header=['ID', 'Name', 'Birth date', 'Weight', 'Height',
                                               'Gender', 'Activity Level'],
                                       data=db.fetch_all(table="users"))

    @staticmethod
    def modify_user():
        logger.info("Modifying user")
        print("Choose what user to modify: ")
        UserActions.list_users()
        while True:
            id_to_modify: str = input("ID of user to modify: ")
            try:
                with DBController() as db:
                    number_of_users: int = len(db.fetch_all(table="users"))
                id_to_modify: int = int(id_to_modify)
                if 1 <= id_to_modify <= number_of_users:
                    break
                else:
                    raise InvalidArgumentError(f"User id must be in range 1 - {number_of_users}")
            except (TypeError, InvalidArgumentError):
                logger.error("Incorrect choice, try again")

        logger.info("Obtaining all the pieces of information about edited ")
        with DBController() as db:
            fetched_info: list = db.fetch_user_by_id(row_id=id_to_modify)[0]
            fields: list = ['user_name', 'date_of_birth', 'weight', 'height', 'gender', 'activity_level']
            default_info: dict = {field: info for field, info in zip(fields, fetched_info)}

        logger.info("Obtaining new info")
        user_name, date_of_birth, weight, height, gender, activity_level = UserActions._obtain_user_info(
            default_values=default_info)

        with DBController() as db:
            db.update_user(name=user_name, date_born=date_of_birth, weight=weight,
                           height=height, gender=gender, activity=ActivityLevel(activity_level),
                           id_to_modify=id_to_modify)
        logger.info("Update: success")
        print("User was updated successfully.")

    @staticmethod
    def delete_user():
        raise NotImplementedError

    @staticmethod
    def export_users():
        raise NotImplementedError



from datetime import date
from os.path import dirname
from sqlite3 import Connection, connect, IntegrityError, Cursor, Error as SQLError
from typing import Any

import sport_tracker
from sport_tracker.common.exceptions import InvalidArgumentError
from sport_tracker.logger import logger
from sport_tracker.model.person import ActivityLevel


class DBController:
    def __init__(self):
        self.db_file: str = f"{dirname(sport_tracker.__file__)}/model/database.db"  # db file is auto-created
        self.connection: Connection
        self.statements = {"UPDATE_USER": "UPDATE users SET 'name'=?, 'born_date'=?, 'weight'=?, 'height'=?, "
                                          "'gender'=?, 'activity_level'=? WHERE id=?",
                           "UPDATE_SPORT": "UPDATE sports SET 'name'=?, 'moving'=? WHERE id=?",
                           "UPDATE_ACTIVITY": "UPDATE activities SET 'sport_id'=?, 'user_id'=?, 'time'=?, "
                                              "'distance'=? WHERE id=?",
                           "INSERT_USER": "INSERT INTO users('name', 'born_date', 'weight', 'height', 'gender',"
                                          " 'activity_level') "
                                          "VALUES (?, ?, ?, ?, ?, ?);",
                           "INSERT_ACTIVITY": "INSERT INTO activities('sport_id', 'user_id', 'time', 'distance') "
                                              "VALUES (?, ?, ?, ?);",
                           "INSERT_SPORT": "INSERT INTO sports('name', 'moving') VALUES (?, ?);",
                           # table names cannot be parametrized, thus separate delete statements
                           "DELETE_USER": "DELETE FROM users WHERE id=?",
                           "DELETE_ACTIVITY": "DELETE FROM activities WHERE id=?",
                           "DELETE_SPORT": "DELETE FROM sports WHERE id=?",
                           # table names cannot be parametrized, thus separate fetch statements
                           "FETCH_ALL_USERS": "SELECT id, name, born_date, weight, height, "
                                              "CASE WHEN gender = 0 then 'male' else 'female' END, activity_level"
                                              " FROM users LIMIT ?;",
                           "FETCH_USER_BY_ID": "SELECT name, born_date, weight, height, "
                                               "CASE WHEN gender = 0 then 'male' else 'female' END, "
                                               "activity_level FROM users WHERE id=?;",
                           "FETCH_ALL_ACTIVITIES": "SELECT * FROM activities LIMIT ?;",
                           "FETCH_ALL_SPORTS": "SELECT id, name, CASE WHEN moving = 1 then 'yes' else 'no' END "
                                               "FROM sports LIMIT ?;",
                           # table names cannot be parametrized, thus separate fetch row id statements
                           "FETCH_ROW_ID_USERS": "SELECT ROWID FROM users WHERE name=?;",
                           "FETCH_ROW_ID_SPORTS": "SELECT ROWID FROM sports WHERE name=?;",
                           }

    def create_connection(self) -> Connection:
        """
        Creates connection to the database file
        :return: established connection
        """
        try:
            return connect(self.db_file)
        except SQLError as e:
            logger.error(str(e))
            exit(1)

    def __enter__(self):
        self.connection = self.create_connection()
        self._populate_db()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.close()
        if exc_val:
            raise

    def _execute(self, statement: str, *args) -> Cursor:
        """
        Executes statement from :statements: dictionary with given parameters
        :param statement: statement from :statements" dictionary to be executed
        :param args:
        :return: cursor from the connection
        """
        if statement not in self.statements:
            raise InvalidArgumentError(f"statement argument must be one of the {self.statements.keys()}")
        c = self.connection.cursor()
        try:
            c.execute(self.statements[statement], args)
        except IntegrityError:
            raise
        except SQLError as e:
            logger.error(str(e))
            self.connection.close()
            raise
        else:
            self.connection.commit()
            return c

    # initial population
    def _populate_db(self):
        with open(f"{dirname(sport_tracker.__file__)}/common/create_tables.sql") as create_tables_sql:
            try:
                tables_creation_statements = create_tables_sql.read()
                for single_table_creation_statement in tables_creation_statements.split(";"):
                    self.connection.execute(single_table_creation_statement)
                self.connection.commit()
            except (FileNotFoundError, SQLError) as e:
                logger.error(str(e))
                raise

    # public insert methods
    def insert_user(self, *, name: str, date_born: date, weight: float, height: int, gender: int,
                    activity: ActivityLevel) -> bool:
        """
        Method inserts new user into database
        :param name: User's name
        :param date_born: date of birth
        :param weight: weight of the user in kg's
        :param height: height of the user in cm's
        :param gender: user's gender, 0 for male, 1 for female
        :param activity: level of activity (see Person class file)
        :return: True if operation was successful, False otherwise
        :raise: error different than IntegrityError which is handled
        """
        try:
            self._execute("INSERT_USER", name, date_born.strftime("%Y-%m-%d"), weight, height, gender, activity.value)
        except IntegrityError as e:
            if "name" in str(e):
                logger.error("User's name already exists in the database")
                return False
            else:
                raise
        else:
            return True

    def insert_activity(self, *, sport_id: int, user_id: int, time: int, distance: int = 0) -> bool:
        """
        Method inserts new general activity into database
        :param sport_id: rowid of the sport from :sports: table
        :param user_id: rowid of the user from :users: table
        :param time: total time spent by doing this activity
        :param distance: total distance done doing this activity, defaults to zero for static activity
        :return: True if the operation was completed successfully
:       """
        self._execute("INSERT_ACTIVITY", sport_id, user_id, time, distance)
        return True

    def insert_moving_activity(self, *, sport_id: int, user_id: int, time: int, distance: int) -> bool:
        """
        Method inserts new moving activity into database
        :param sport_id: rowid of the sport from :sports: table
        :param user_id: rowid of the user from :users: table
        :param time: total time spent by doing this activity
        :param distance: total distance done doing this activity
        :return: True if the operation was completed successfully
        """
        self.insert_activity(sport_id=sport_id, user_id=user_id, time=time, distance=distance)
        return True

    def insert_sport(self, *, name: str, moving: bool) -> bool:
        """
        Method inserts new sport into database
        :param name: name of the sport to be inserted
        :param moving: boolean: designates moving, i.e. distance based sport
        :return: True if operation was successful, False otherwise
        :raise: error different than IntegrityError which is handled
        """
        try:
            self._execute("INSERT_SPORT", name, 1 if moving else 0)
        except IntegrityError as e:
            if 'name' in str(e):
                logger.error(f"Sport named {name} already exists in the database")
            else:
                raise
        else:
            return True

    # public delete methods
    def delete_user(self, *, row_id: int):
        # TODO: Implement
        pass

    def delete_activity(self, *, row_id: int):
        # TODO: Implement
        pass

    def delete_sport(self, *, row_id: int):
        # TODO: Implement
        pass

    # fetch methods
    def fetch_row_id(self, *, table: str, column: str = 'name', value: Any) -> int:
        result_cursor: Cursor = self._execute(f"FETCH_ROW_ID_{table.upper()}", value)
        result_list: list = result_cursor.fetchall()
        if not len(result_list):
            raise IntegrityError(f"No value for '{value}' in column '{column}' of table '{table}' found!")
        if len(result_list) == 1:
            return result_list[0][0]
        else:
            raise IntegrityError(f"Multiple values for '{value}' in column '{column}' of table '{table}' found!")

    def fetch_all(self, *, table: str, limit: int = 50) -> list:
        result_cursor: Cursor = self._execute(f"FETCH_ALL_{table.upper()}", limit)
        return result_cursor.fetchall()

    def fetch_user_by_id(self, *, row_id: int) -> list:
        result_cursor: Cursor = self._execute("FETCH_USER_BY_ID", row_id)
        return result_cursor.fetchall()

    # update methods
    def update_user(self, *, name: str, date_born: date, weight: float, height: int, gender: int,
                    activity: ActivityLevel, id_to_modify: int) -> bool:
        self._execute("UPDATE_USER", name, date_born.strftime("%Y-%m-%d"), weight, height, gender,
                      activity.value, id_to_modify)
        return True

    def update_sport(self, *, name: str, moving: bool, id_to_modify: int) -> bool:
        self._execute("UPDATE_SPORT", name, 1 if moving else 0, id_to_modify)
        return True

    def update_activity(self, *, sport_id: int, user_id: int, time: int, distance: int = 0, id_to_modify: int) -> bool:
        self._execute("UPDATE_ACTIVITY", sport_id, user_id, time, distance, id_to_modify)
        return True
